<?php
/**
 * dbupgrade.php
 *
 * Database upgrade for the main application and all 3rd party modules/applications.
 * Refer to cmsupgradebase.php for detail of the upgrade system.
 *
 * @author $Author: dtong $
 * @version $Id: dbupgrade.php,v 1.16 2012/02/09 07:23:00 dtong Exp $
 * @copyright Copyright (c) 2011, Tiller Software Co., Ltd.
*/

class Dbupgrade extends CI_Controller {
	const TIME_LIMIT = 600;
	const DB_INIT_FILE = 'db_init.sql';
	const DB_INIT_FILE_NO_EXT = 'db_init';

	// Statuses for the install process.
	const INSTALL_ERROR = '<!-- STATUS_ERROR -->';
	const INSTALL_OK = '<!-- STATUS_OK -->';
	const INSTALL_DONE = '<!-- STATUS_DONE -->';

	//The TODCM main DB upgrade folder
	private $dir_script = '';
	private $dir_3p;
	private $current_version = '';//store db current version.
	private $current_module='';
	private $current_module_path='';
	private $current_sql = '';
	private $last_upgraded_version = '';

	function __construct() {
		parent::__construct();
		osa_load_lang('admin');
		$this->load->model('admin/dbupgrade_model');
		$this->dir_script = FCPATH . 'install/upgrade/db/';
		$this->dir_3p = APPPATH . 'third_party/';
	}

	function index() {
		$this->access_model->checkaccess_any('admin_system_management');
		$this->_update_3p_config();
		$has_scripts = $this->_all_upgrade_scripts();
		$data = new stdClass();
		$data->admin_dbupgrade_currentversion = lang('admin_dbupgrade_currentversion');
		$data->admin_dbupgrade_title = lang('admin_dbupgrade_title');
		$data->current_ver = $this->dbupgrade_model->get_current_version();
		//if $this->script have data will do upgrade.
		if ( $has_scripts ) {
			$data->admin_dbupgrade_warn = lang('admin_dbupgrade_warn');
			$data->admin_dbupgrade_listofupgrade = lang('admin_dbupgrade_listofupgrade');
			$data->admin_dbupgrade_iframe_headtitle = lang('admin_dbupgrade_iframe_headtitle');
			$data->admin_dbupgrade_iframe_text = lang('admin_dbupgrade_iframe_text');
			$data->script_ver = $has_scripts;
			$ret = $this->load->view('admin/dbupgrade_warning_view', $data, TRUE);
		}
		else {
			$data->admin_dbupgrade_db_latest = lang('admin_dbupgrade_db_latest');
			$ret = $this->load->view('admin/dbupgrade_view', $data, TRUE);
		}
		echo $ret;
		// Update 3p js, css and image.
		if ( osa_3p_available() &&
			 ($jsimg_obj = $this->cms_3p->load_lib('Cms_prepare_js_image', CMS_3P_FOLDER_COMMON)) ) {
			$jsimg_obj->run();
		}
	}

	// This is for the application install to check for readiness of the install and key checking.
	// TODO: We piggy back on the config encryption_key but should make a new subsystem to
	//			handle temporary access keys.
	function prepare($key=FALSE) {
		// Do not put access check in this function.
		if ( !$this->_valid_key($key) ) {
			// Invalid key
			echo self::INSTALL_ERROR;
			exit;
		}
		// Have to run this so we have the necessary 3p configs before upgrading.
		$this->_update_3p_config(TRUE);
		// Everything is ready and the caller should continue to upgrade 3rd party apps.
		$scripts = $this->_next_upgrade_app();
		// Update 3p js, css and image. This is like making the source ready so
		// happens here before TODCM and 3P upgrades.
		if ( osa_3p_available() &&
			 ($jsimg_obj = $this->cms_3p->load_lib('Cms_prepare_js_image', CMS_3P_FOLDER_COMMON)) ) {
			// This can do an exit so make sure to put it last.
			if ( !$jsimg_obj->run(TRUE) ) {
				echo self::INSTALL_ERROR;
				exit;
			}
		}
		if ( $scripts ) {
			echo self::INSTALL_OK;
		}
		else {
			echo self::INSTALL_DONE;
		}
		exit;
	}

	// $key is used for during initial installation only
	function upgrade($key=FALSE) {
		if ( !$this->_valid_key($key) ) {
			// Check access if we don't have a valid key
			$this->access_model->checkaccess_any('admin_system_management');
		}
		set_time_limit(self::TIME_LIMIT);
		ob_implicit_flush(true);
		ob_end_flush();
		$upgraded = FALSE;
		if ( !$key ) {
			$data = new stdClass();
			$data->admin_dbupgrade_iframe_title = lang('admin_dbupgrade_iframe_title');
			echo $this->load->view('admin/dbupgrade_upgrading_head_view', $data, true);
		}
		while ($scripts = $this->_next_upgrade_app()) {
			if ( $key && $upgraded ) {
				echo self::INSTALL_OK;
				exit;
			}
			$ret = $this->_upgrade_dir($scripts, $this->current_module_path);
			$system = $this->_system_name($this->current_module_path);
			if ( $system === FALSE ) {
				$ret = FALSE;
			}
			if ( !$ret ) {
				if ( $key ) {
					echo self::INSTALL_ERROR;
				}
				echo '<br /><br /><div style="color:red;font-size:12px;font-weight:bold">' . lang('admin_dbupgrade_failed_msg1');
				echo "<br /><br />" . lang('admin_dbupgrade_sql') . "{$this->current_sql}<br /><br />" .
						lang('admin_dbupgrade_error_num') . "{$this->dbupgrade_model->error_number}<br /><br />" .
						lang('admin_dbupgrade_error_msg') . "{$this->dbupgrade_model->error_message}<br /><br /><br/>" .
						lang('admin_dbupgrade_failed_msg2');
				echo '</div>';
				if ( !$key ) {
					echo $this->load->view('admin/dbupgrade_upgrading_foot_view', array('failed_upgrade'=>TRUE), true);
				}
				exit;
			}
			if ( $this->last_upgraded_version != self::DB_INIT_FILE_NO_EXT ) {
				// Update the version # in the DB for each system.
				// This is skipped if the script is db_init since we can't determine the version
				// from the db_init.sql file name therfore all db_inti.sal script has to
				// update the version # from within the script.
				$this->dbupgrade_model->update_version($this->last_upgraded_version, $system);
			}
			$upgraded = TRUE;
			if ( !$key ) {
				// Force browser output;
				// echo str_repeat(' ', 32);
				@ob_end_flush();
			}
		}
		if ( $key ) {
			// if we get here from within the install process then this means no more app to upgrade
			echo self::INSTALL_DONE;
			exit;
		}
		// We always update the cms_db_version based on the current DB version.
		// A new install will insert cms_db_version to the app_config table so no need for the following
		// when installing.
		$main_version = $this->dbupgrade_model->get_current_version();
		osa_config_update2db('cms_db_version', $main_version);
		$this->_clean_cache();
		echo '<br />** ' . lang('admin_dbupgrade_success_msg') . ' **<br />';
		echo $this->load->view('admin/dbupgrade_upgrading_foot_view', FALSE, true);
		exit;
	}

	// Gets all the upgrade scripts that need to be run for all applications.
	private function _all_upgrade_scripts($showpath=TRUE) {
		$ret = array();
		$count = 0;
		while ($scripts = $this->_next_upgrade_app()) {
			if ( $showpath ) {
				foreach ( $scripts as $key=>$valye) {
					if ( $scripts[$key] == self::DB_INIT_FILE ) {
						$version = lang('admin_dbupgrade_todcm_dbinit');
					}
					else {
						$version = osa_str_extract($scripts[$key], '.', TRUE, FALSE);
					}
					// replace all \ with /, necessary for windows systems
					$path = str_replace('\\', '/', $this->current_module_path);
					$scripts[$key] = "$version ({$path})";
				}
			}
			$ret = array_merge($ret, $scripts);
		}
		return $ret;
	}

	private function _real_upgrade_php($upgrade_version, $db_dir, $action) {
		$tmp_dbdir = $db_dir;
		osa_str_addslash($tmp_dbdir);
		$file = "{$tmp_dbdir}{$upgrade_version}_{$action}.php";
		if ( !file_exists($file) ) {
			return TRUE;
		}
		$system = $this->_system_name($db_dir);
		if ( $this->_is_main_system($system) ) {
			$class = "upgrade_{$upgrade_version}_{$action}";
		}
		else {
			$class = "upgrade_{$system}_{$upgrade_version}_{$action}";
		}
		// Can not have dots in class name so make them '_'.
		$class = str_replace('.', '_', $class);
		$ret = include_once($file);
		$is_class = class_exists($class);
		if ( !$ret || !$is_class ) {
			echo '<font color="red">' . sprintf(lang('admin_dbupgrade_failed_php'), $file) . '</font>';
			return FALSE;
		}
		$obj = new $class;
		if ( property_exists($obj, 'error_msg') && $obj->error_msg ) {
			echo "<font color=\"red\">{$file}: {$obj->error_msg}</font>";
			return FALSE;
		}
		return TRUE;
	}

	// Runs the pre SQL PHP file
	private function _pre_upgrade_php($upgrade_version, $db_dir) {
		return $this->_real_upgrade_php($upgrade_version, $db_dir, 'pre');
	}

	// Runs the post SQL PHP file
	private function _post_upgrade_php($upgrade_version, $db_dir) {
		return $this->_real_upgrade_php($upgrade_version, $db_dir, 'post');
	}

	// Upgrades all the necessary scripts in one DB folder
	private function _upgrade_dir($scripts, $db_dir, $system=NULL) {
		//$has_scripts = $this->_get_scripts($db_dir, $system);
		// $data = new stdClass();
		// $data->admin_dbupgrade_iframe_title = lang('admin_dbupgrade_iframe_title');
		// echo $this->load->view('admin/dbupgrade_upgrading_head_view', $data, true);
		$this->flush_buffer();
		//assign for display.
		$current_version = $this->current_version;
		osa_str_addslash($db_dir);

		//query newer script.
		foreach ( $scripts as $script_name ){
			$upgrade_version = basename($script_name, '.sql');
			if ( $upgrade_version != self::DB_INIT_FILE_NO_EXT ) {
				echo lang('admin_dbupgrade_upgraded_from') . " $current_version " . lang('admin_dbupgrade_upgraded_to') .
					  " <span style=\"color:red;\">{$upgrade_version} ({$db_dir})</span><br/>";
			}
			else {
				echo lang('admin_dbupgrade_todcm_dbinit') . " ({$db_dir})<br/>";
			}
			$this->flush_buffer();
			$sql = osa_next_sql(realpath($db_dir . $script_name));
			// $count = 0;
			$line = 1;
			if ( !$this->_pre_upgrade_php($upgrade_version, $db_dir) ) {
				$this->current_sql = '';
				return FALSE;
			}
			while ($sql) {
				$this->current_sql = $sql;
				//echo $line.' : '.$sql.'<br />';//use for check sql that will query.
				$ret = $this->_todcm_query($sql);

				if ( !$ret ) {
					return FALSE;
				}
				/*
				if ( $count > 32 ) {
					echo '<br />';
					$count = 0;
				}
				*/
				echo '. . ';
				$this->flush_buffer();
				$sql = osa_next_sql();
				// $count++;
				$line++;
			}
			if ( !$this->_post_upgrade_php($upgrade_version, $db_dir) ) {
				$this->current_sql = '';
				return FALSE;
			}
			echo '<hr />';
			$this->flush_buffer();
			//change current version to upgraded version for display.
			$current_version = $upgrade_version;
		}
		$this->last_upgraded_version = $upgrade_version;
		return TRUE;
	}

	/*
	 * Get the scripts that need to be run in one DB folder.
	 * Return false if no newer script.
	 */
	private function _get_scripts($folder) {
		//get database current version.
		$system = $this->_system_name($folder);
		if ( !file_exists($folder) ) {
			if ( $this->_is_main_system($system) ) {
				// The TODCM DB upgrade folder is missing
				echo osa_ajaxmsg(lang('admin_dbupgrade_todcm_dbfolder_missing'));
				exit;
			}
			return FALSE;
		}
		$this->current_version = $this->dbupgrade_model->get_current_version($system);
		if ( !$this->current_version ) {
			if ( $this->_is_main_system($system) ) {
				// Main TODCM should always have the version # so print error
				echo osa_ajaxmsg(lang('admin_dbupgrade_missing_version'));
				exit;
			}
			// Current version is missing in DB so this must be a new 3rd party app.
			$init_file = $folder;
			osa_str_addslash($init_file);
			$init_file .= self::DB_INIT_FILE;
			if ( file_exists($init_file) ) {
				// We have an init file so ignore other files and return it
				return array(self::DB_INIT_FILE);
			}
			// No current system version and no init file so just return false
			// This implies there should be always an init file if there are version upgrade file.
			// It would be a misconfiguration if init file is missing and version file(s) are present.
			return FALSE;
		}
		$upgrade_scripts = $this->_get_upgrade_scripts($folder);
		// if( empty($this->current_version) || empty($script_versions) || $this->current_version == 'APPLICATION_VERSION_NUMBER' ){
		if ( empty($this->current_version) || empty($upgrade_scripts) ) {
			return FALSE;
		}
		$db_current_version = $this->current_version . '.sql';//need this to correct comparing.
		$scripts = array();
		foreach( $upgrade_scripts as $script_name ) {
			//collect newer script for upgrade.
			if ( $this->_compare_version($db_current_version, $script_name) ) {
				$scripts[] = $script_name;
			}
		}
		if ( empty($scripts) ) {
			return FALSE;
		}
		return $scripts;
	}

	// Each call to this function will return the next avaiable upgrade scripts belonging to one application
	private function _next_upgrade_app() {
		static $count = 0;
		$this->current_module_path = '';
		if ( $count === NULL ) {
			return FALSE;
		}
		// For main application
		if ( $count == 0 ) {
			$ret = $this->_get_scripts($this->dir_script);
			if ( $ret ) {
				$count++;
				$this->current_module = NULL;
				$this->current_module_path = $this->dir_script;
				return $ret;
			}
		}
		// For all 3rd party applications
		if ( osa_3p_available() ) {
			// We start from the very begining eery time this is called
			$tmp_count = 1;
			// This has all the avaiable modules
			$tp_apps = $this->config->item('cms_3p_modules');
			// Lopp thru the applications and find the next upgrade script if there is one
			foreach ( $tp_apps as $group_key=>$group ) {
				// This is the group level
				if ( is_object($group) ) {
					foreach ($group as $app_key=>$app_path) {
						// This is the application level
						if ( $tmp_count >= $count ) {
							$path = "{$this->dir_3p}{$app_path}/db";
							$ret = $this->_get_scripts($path);
							if ( $ret ) {
								// We increment $count so we won't go into the same script in the next iteration
								$count = ++$tmp_count;
								$this->current_module = $app_key;
								$this->current_module_path = $path;
								return $ret;
							}
						}
						$tmp_count++;
					}
				}
			}
		}
		// This means we reached the end of the upgrade scripts
		$count = NULL;
		return FALSE;
	}

	// Get all the existing scripts in one DB folder. The version compare is done elsewhere.
	private function _get_upgrade_scripts($db_dir) {
		$script_names = array();
		$hand = @opendir($db_dir);
		if ( empty($hand) ) {
			return FALSE;
		}
		while( $file = @readdir($hand) ) {
			//collect only .sql file names and skip db_init.sql
			if( strripos($file, '.sql') && $file != self::DB_INIT_FILE ) {
				$script_names[] = $file;
			}
		}
		closedir($hand);
		usort($script_names, 'strnatcasecmp');
		return $script_names;
	}

	private function _is_main_system($system) {
		return ( $system === NULL );
	}

	// compare current database with script version.
	private function _compare_version($current_ver, $script_ver) {
		return (strnatcasecmp($current_ver, $script_ver) < 0);
	}

	// Gets the system name based on the upgrade DB path
	private function _system_name($path) {
		if ( empty($path) ) {
			echo sprintf(lang('admin_dbupgrade_apppath_error'), $path);
			return FALSE;
		}
		if ( stripos($path, 'third_party') === FALSE ) {
			$system = NULL;
		}
		else {
			$system = basename(dirname($path));
			if ( !$system ) {
				echo sprintf(lang('admin_dbupgrade_apppath_error'), $path);
				return FALSE;
			}
		}
		return $system;
	}

	// Run a single query
	private function _todcm_query($sql) {
		$sql = str_replace(TODCM_APPLICATION_VERSION_NUMBER, CMS_SOURCE_VERSION, $sql);
		return $this->dbupgrade_model->run_sql(trim($sql));
	}

	private function flush_buffer() {
		// We don't want any buffering.
		// CI can turn on buffer so just end the buffer and flush even if there isn't one.
		@ob_end_flush();
	}

	private function _update_3p_config($bootstrap=FALSE) {
		// Update the 3rd party configs
		$params = array();
		if ( $bootstrap ) {
			$params['bootstrap'] = TRUE;
		}
		$this->load->library('cms/Cms_3p', $params);
		$this->cms_3p->update_config();
	}

	private function _valid_key($key) {
		return ( !empty($key) && strcasecmp($key, $this->config->item('encryption_key') ===0) );
	}

	// Clean the presistent storage since they can be inconsistent after a upgrade. This is a 3p only requirement.
	// We always run DB upgrade script even if there is no actual DB changes so we can be sure this gets run on every upgrade.
	// This also refreshes the template object caches.
	private function _clean_cache() {
		$this->load->library('admin/Admin_common');
		$this->admin_common->cleanup();
	}
}