<?php
/**
 * cms_dyn_config
 *
 * Dynamic configuration file creation and update. This will strip all / * * / comments but // is ok.
 * All object and array date will be ignored. Make sure to esacpe ' in your strings.
 *
 * If you are passing in a varaible like $config['var'] then do not ever use $config["var"] else where.
 * They are treated as 2 separate variables in this class. You should always use single quotes unless there is a rason not to.
 *
 * @author $Author: dtong $
 * @version $Id: cms_dyn_config.php,v 1.2 2012/02/27 04:32:13 dtong Exp $
 * @copyright Copyright (c) 2012, Tiller Software Co., Ltd.
*/

class cms_dyn_config {
	const COMMENT_TEXT = 'Generated by %s. DO NOT MODIFY.';
	const NEW_LINE = "\n";
	const PHP_TAG = '<?php';
	const IS_COMMENT = '__COMMENT_LINE__';
	const TMP_EXTENSION = '._dconf_';

	private $this_filename;

	function __construct() {
		$this->this_filename = basename(__FILE__);
	}

	function put($filepath, $data, $title_text=TRUE) {
		if ( !$filepath || empty($data) || !is_array($data) ) {
			return FALSE;
		}
		if ( !$title_text ) {
			$buffer = self::PHP_TAG . self::NEW_LINE;
		}
		else {
			$buffer = self::PHP_TAG . self::NEW_LINE . '// ' . sprintf(self::COMMENT_TEXT, $this->this_filename) . self::NEW_LINE;
		}
		// Merge existing variables with the new ones.
		$data = $this->merge($filepath, $data);
		if ( empty($data) ) {
			return FALSE;
		}
		foreach ($data as $key => $value) {
			if ( is_array($value) ) {
				continue;
			}
			if ( is_object($value) ) {
				continue;
			}
			if ( $value === FALSE ) {
				$value = 'false';
			}
			elseif ( $value === TRUE ) {
				$value = 'true';
			}
			elseif ( $value === NULL ) {
				$value = 'null';
			}
			elseif ( is_numeric($key) ) {
				// Comment lines have numeric index and we keep the comments
				$buffer .= $value . self::NEW_LINE;
				continue;
			}
			else {
				$value = "'{$value}'";
			}
			$buffer .= $key . '=' . $value . ';' . self::NEW_LINE;
		}
		// Use rename/move to ensure no concurrent access issues.
		if ( !(@file_put_contents($filepath . self::TMP_EXTENSION, $buffer)) ) {
			return FALSE;
		}
		if ( !(@rename($filepath . self::TMP_EXTENSION, $filepath)) ) {
			osa_log(__METHOD__, 'Failed to rename file.', $filepath . self::TMP_EXTENSION, $filepath);
			@unlink($filepath . self::TMP_EXTENSION);
			return FALSE;
		}
		return TRUE;
	}

	 private function merge($filepath, $data) {
		if ( !file_exists($filepath) ) {
			return $data;
		}
		$current_data = $this->get($filepath);
		if ( empty($current_data) ) {
			return $data;
		}
		$ret = array();
		foreach($current_data as $key=>$value) {
			if ( array_key_exists($key, $data) ) {
				$ret[$key] = $data[$key];
				unset($data[$key]);
			}
			else {
				$ret[$key] = $value;
			}
		}
		if ( empty($data) ) {
			return $ret;
		}
		return $ret + $data;
	}

	private function get($filepath) {
		if ( ($buffer = file_get_contents($filepath)) === FALSE ) {
			return array();
		}
		$data = array();
		$buffer = str_replace("\r\n", "\n", $buffer);
		$buffer = str_replace("\r", "\n", $buffer);
		$buffer = explode("\n", $buffer);
		foreach ($buffer as $line) {
			$line = trim($line);
			if ( strpos($line, '//') === 0 && strpos($line, $this->this_filename) === FALSE ) {
				// We don't keep the comment cam from this file since it will get generated again.
				$data[] = $line;
			}
			elseif( ($dollar_sign = strpos($line, '$')) === 0 && ($equal = strpos($line, '=')) !== FALSE &&
						($semi_colon = strpos($line, ';')) !== FALSE && $semi_colon > $equal ) {
				$length = $semi_colon - $equal - 1;
				if ( $length <= 0 ) {
					continue;
				}
				$value = trim(substr($line, $equal+1, $semi_colon - $equal - 1));
				// Easier to keep the original value using eval. This class is not used often so ok to use eval.
				if ( @eval("\$value={$value};") === FALSE ) {
					continue;
				}
				$var = trim(substr($line, 0, $equal));
				$data[$var] = $value;
			}
			// We ignore all the rest of the stuff like / * * / comments
		}
		return $data;
	}

}